import bpy, math
from ast import parse

from .. rpy import RpyInterpreter

NAMESPACE = {
    'bpy': bpy,
    'acos': math.acos,
    'acosh': math.acosh,
    'asin': math.asin,
    'asinh': math.asinh,
    'atan': math.atan,
    'atan2': math.atan2,
    'atanh': math.atanh,
    'cbrt': math.cbrt,
    'ceil': math.ceil,
    'comb': math.comb,
    'copysign': math.copysign,
    'cos': math.cos,
    'cosh': math.cosh,
    'degrees': math.degrees,
    'dist': math.dist,
    'e': math.e,
    'erf': math.erf,
    'erfc': math.erfc,
    'exp': math.exp,
    'exp2': math.exp2,
    'expm1': math.expm1,
    'fabs': math.fabs,
    'factorial': math.factorial,
    'floor': math.floor,
    'fmod': math.fmod,
    'frexp': math.frexp,
    'fsum': math.fsum,
    'gamma': math.gamma,
    'gcd': math.gcd,
    'hypot': math.hypot,
    'inf': math.inf,
    'isclose': math.isclose,
    'isfinite': math.isfinite,
    'isinf': math.isinf,
    'isnan': math.isnan,
    'isqrt': math.isqrt,
    'lcm': math.lcm,
    'ldexp': math.ldexp,
    'lgamma': math.lgamma,
    'log': math.log,
    'log10': math.log10,
    'log1p': math.log1p,
    'log2': math.log2,
    'modf': math.modf,
    'nan': math.nan,
    'nextafter': math.nextafter,
    'perm': math.perm,
    'pi': math.pi,
    'pow': math.pow,
    'prod': math.prod,
    'radians': math.radians,
    'remainder': math.remainder,
    'sin': math.sin,
    'sinh': math.sinh,
    'sqrt': math.sqrt,
    'tan': math.tan,
    'tanh': math.tanh,
    'tau': math.tau,
    'trunc': math.trunc,
    'ulp': math.ulp,

    'abs': abs,
    'all': all,
    'any': any,
    'ascii': ascii,
    'bin': bin,
    'callable': callable,
    'chr': chr,
    'dir': dir,
    'divmod': divmod,
    'format': format,
    'getattr': getattr,
    'hasattr': hasattr,
    'hash': hash,
    'hex': hex,
    'id': id,
    'isinstance': isinstance,
    'issubclass': issubclass,
    'iter': iter,
    'len': len,
    'max': max,
    'min': min,
    'next': next,
    'oct': oct,
    'ord': ord,
    'pow': pow,
    'repr': repr,
    'round': round,
    'sorted': sorted,
    'sum': sum,
    'bool': bool,
    'memoryview': memoryview,
    'bytearray': bytearray,
    'bytes': bytes,
    'complex': complex,
    'dict': dict,
    'enumerate': enumerate,
    'filter': filter,
    'float': float,
    'frozenset': frozenset,
    'property': property,
    'int': int,
    'list': list,
    'map': map,
    'range': range,
    'reversed': reversed,
    'set': set,
    'slice': slice,
    'str': str,
    'tuple': tuple,
    'type': type,
    'zip': zip,
}
CALLABLE = {
    math.acos,
    math.acosh,
    math.asin,
    math.asinh,
    math.atan,
    math.atan2,
    math.atanh,
    math.cbrt,
    math.ceil,
    math.comb,
    math.copysign,
    math.cos,
    math.cosh,
    math.degrees,
    math.dist,
    math.erf,
    math.erfc,
    math.exp,
    math.exp2,
    math.expm1,
    math.fabs,
    math.factorial,
    math.floor,
    math.fmod,
    math.frexp,
    math.fsum,
    math.gamma,
    math.gcd,
    math.hypot,
    math.isclose,
    math.isfinite,
    math.isinf,
    math.isnan,
    math.isqrt,
    math.lcm,
    math.ldexp,
    math.lgamma,
    math.log,
    math.log10,
    math.log1p,
    math.log2,
    math.modf,
    math.nextafter,
    math.perm,
    math.pow,
    math.prod,
    math.radians,
    math.remainder,
    math.sin,
    math.sinh,
    math.sqrt,
    math.tan,
    math.tanh,
    math.trunc,
    math.ulp,
    abs,
    all,
    any,
    ascii,
    bin,
    callable,
    chr,
    dir,
    divmod,
    format,
    getattr,
    hasattr,
    hash,
    hex,
    id,
    isinstance,
    issubclass,
    iter,
    len,
    max,
    min,
    next,
    oct,
    ord,
    repr,
    round,
    sorted,
    sum,
    bool,
    memoryview,
    bytearray,
    bytes,
    complex,
    dict,
    enumerate,
    filter,
    float,
    frozenset,
    property,
    int,
    list,
    map,
    range,
    reversed,
    set,
    slice,
    str,
    tuple,
    type,
    zip,
}

interpreter = RpyInterpreter()
interpreter.init()
interpreter.attrs_visit = None

def r_filtexp_result(string_expressioN, find, filter_function):
    def each(filter_fn=lambda *k, **kw: True, **kw):
        return find('!', filter_fn, **kw)
    def union(*ss, filter_fn=filter_function, **kw):
        results = set()
        for s in ss:
            results |= find(s, filter_fn, **kw)
        return results
    def intersection(*ss, filter_fn=filter_function, **kw):
        if ss:
            results = find(ss[0], filter_fn, **kw)
            for s in ss[1 : ]:
                results &= find(s, filter_fn, **kw)
            return results
        return set()
    def difference(*ss, filter_fn=filter_function, **kw):
        if ss:
            results = find(ss[0], filter_fn, **kw)
            for s in ss[1 : ]:
                results -= find(s, filter_fn, **kw)
            return results
        return set()

    fns = {
        'find': find,
        'filter_function': filter_function,
        'each': each,
        'union': union,
        'intersection': intersection,
        'difference': difference,
    }
    interpreter.fns_call = CALLABLE | set(fns.values())
    interpreter.env = NAMESPACE | fns

    try:
        return True, interpreter.visit(parse(f'_out = {string_expressioN}'))['_out']
    except Exception as ex:
        return False, str(ex)
    #|
